-- srb2persona font drawing.
-- this script is not re-usable without permission.

if V_drawString and V_drawLetter and V_fontKerning return end	-- srb2persona string libs already loaded, do not process.

local ascii_03d = {}
for i = 0, 255	-- generate ascii %03d table so we don't have to do string.format everytime, this function is very slow:
	ascii_03d[i] = string.format("%03d", tostring(i))
end

-- font kerning: shift the letter width by a set number of pixels if that helps make em look better
rawset(_G, "V_fontKerning", {})

rawset(_G, "V_drawLetter", function(v, x, y, str, prefix, flags, fullscale)
	if not (v and tonumber(x)~=nil and tonumber(y)~=nil and str and prefix) return end
	str = tostring(str)
	if str:len() > 1
		return
	end

	if not v.patchExists(prefix..ascii_03d[str:byte()]) return end	-- patch doesn't exist, don't draw
	v.drawScaled(x, y, fullscale and FRACUNIT or FRACUNIT/2, v.cachePatch(prefix..ascii_03d[str:byte()]), flags)
end)

local function V_getKerning(prefix, char)
	return V_fontKerning[prefix] and V_fontKerning[prefix][char] or 0
end

rawset(_G, "V_stringWidth", function(v, str, prefix, flags, fullscale)
	if not (v and str and prefix) return 0 end
	str = tostring(str)

	local space, div, kern = 8, 2, 1
	if fullscale
		space, div, kern = 16, 1, 2
	end

	local xoffs = 0

	for i = 1, str:len()
		local s = str:sub(i, i)
		if not v.patchExists(prefix..ascii_03d[s:byte()]) -- patch doesn't exist, assume it's a space:
			xoffs = $+space
			continue
		end
		local pp = v.cachePatch(prefix..ascii_03d[s:byte()])
		xoffs = $+pp.width/div + V_getKerning(prefix, char)*kern
	end
	return xoffs
end)

rawset(_G, "V_drawString", function(v, x, y, str, prefix, flags, alignment, fullscale)
	if not (v and tonumber(x)~=nil and tonumber(y)~=nil and str and prefix) return end
	str = tostring(str)

	local space, div, kerningmult = 8, 2, 1
	if fullscale
		space, div, kerningmult = 16, 1, 2
	end

	local xoffs = 0
	local yoffs = 0

	-- alignment coordinate adjustment:
	if alignment == "right"
		x = $ - V_stringWidth(v, str, prefix, flags, fullscale)
	elseif alignment == "center"
		x = $ - V_stringWidth(v, str, prefix, flags, fullscale)/2
	end

	for i = 1, str:len()
		local s = str:sub(i, i)
		if s == "\n"	-- line break
			yoffs = $+space
			xoffs = 0
			continue
		end
		if not v.patchExists(prefix..ascii_03d[s:byte()]) -- patch doesn't exist, assume it's a space:
			xoffs = $+space
			continue
		end

		V_drawLetter(v, (x+xoffs)<<FRACBITS, (y+yoffs)<<FRACBITS, s, prefix, flags, fullscale)	-- letter
		local pp = v.cachePatch(prefix..ascii_03d[s:byte()])
		xoffs = $+ pp.width/div + V_getKerning(prefix, s)*kerningmult
	end
end)