-- It's time to wipe out some players

local wver = "V4.1"
local MAXZONEHP = 100

local startzone = CV_RegisterVar({
	name = "z_startzone",
	defaultvalue = "1",
	flags = CV_NETVAR,
	possiblevalue = CV_Natural
})

local starthp = CV_RegisterVar({
	name = "z_starthp",
	defaultvalue = "100",
	flags = CV_NETVAR,
	possiblevalue = CV_Natural,
})

local haszone = CV_RegisterVar({
	name = "z_zonemode",
	defaultvalue = "On",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

local hasbattle = CV_RegisterVar({
	name = "z_zonebattle",
	defaultvalue = "On",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

local forcebattle = CV_RegisterVar({
	name = "z_forcezonebattle",
	defaultvalue = "Off",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

local zonetarget = CV_RegisterVar({
	name = "z_zonebattletarget",
	defaultvalue = "35",
	flags = CV_NETVAR,
	possiblevalue = CV_Natural
})

local hudshake = CV_RegisterVar({
	name = "z_hudshake",
	defaultvalue = "On",
	possiblevalue = CV_OnOff
})

local zoneon -- do we have wipezones enabled
local btlzone -- are we playing zone battle?
local dynzone = zonetarget.value	-- dynamic zone target
local battlefinish = 16				-- used for rankings by setting the zone number absurdly high when you're done

local displayplayer	-- used for earthquake on damage

local starttime = 6*TICRATE + (3*TICRATE/4)
local zoneincrement = FRACUNIT/200	-- how much speed we gain per frame
local zonemsg = {}	-- zone message. ("PERFECT LAP", "COURSE RESET" ...)

local battlemaxboost = FRACUNIT*8
local battlemaxshield = 100

-- send contextual zone message
local function z_sendmsg(p,msg)
	if p and p~=displayplayer return end
	if splitscreen return end
	local t = {msg, TICRATE*5}
	zonemsg[#zonemsg+1] = t
end

-- get decimal point from a fixed value
local function z_getdec(val)
	if not val return "0.0" end
	local full = val/FRACUNIT
	local dec = val%FRACUNIT *10 / FRACUNIT
	return full.."."..dec
end


local zonenames = {
[0] = "Sub-Venom",
[2] = "Venom",
[3] = "Sub-Flash",
[5] = "Flash",
[7] = "Sub-Rapier",
[12] = "Rapier",
[16] = "Sub Phantom",
[20] = "Phantom",
[27] = "Super Phantom",
[35] = "Zen",
[42] = "Super Zen",
[50] = "Subsonic",
[60] = "Mach I",
[75] = "Supersonic",
[690] = "thicc thighs",
[691] = "save lives",
[692] = "thigh highs",
[693] = "are my demise",
[6969] = "Kanade best waifu",
[6970] = "hey yuri is p good",
[32767] = "Manic low speed"
}

-- these play when you reach the specified zone
local zonesfx = {
	[2] = "sfx_zon1",
	[3] = "sfx_zon2",
	[5] = "sfx_zon3",
	[7] = "sfx_zon4",
	[12] = "sfx_zon5",
	[16] = "sfx_zon6",
	[20] = "sfx_zon7",
	[27] = "sfx_zon8",
	[35] = "sfx_zon9",
	[42] = "sfx_zon10",
	[50] = "sfx_zon11",
	[60] = "sfx_zon12",
	[75] = "sfx_zon13",
}

do
	for i = 1, 75
		if zonesfx[i]
			freeslot(zonesfx[i])
		end
	end
end
freeslot("sfx_zwaa", "sfx_zelmn", "sfx_ztime")

-- get zone name relative to where we are
local function z_getzonename(n)

	local name
	for i = 0, min(75, n)	-- don't bother going above 75, there's nothing beyond that
		if zonenames[i]
			name = zonenames[i]
		end
	end
	return name or ""
end

-- kill player and inflict permanent DNF
local function z_blowupplayer(p)
	if not (p and p.mo) or p.spectator return end
	if p.blown return end

	p.pflags = $|PF_TIMEOVER
	p.lives = 0
	p.blown = true
	P_DamageMobj(p.mo, nil, nil, 10000)
	p.blown = nil
	P_FlashPal(p, 1, 10)
	S_ChangeMusic("KRFAIL", true, p)
end

local function z_fillzonegauge(mo)
	-- generate shield and boost
	-- the more we had, the more we get.
	mo.player.zbattleshield = min(battlemaxshield, $ + ($*3/16) + (battlemaxshield*2/17))
	mo.player.zbattleboost = min(battlemaxboost, $ + ($*3/16) + (battlemaxboost*2/17))
end


-- this is a hack to prevent diagonal springs from screwing us over.
addHook("MobjMoveCollide", function(mo, s)
	if not (zoneon) then return end

	if (s and s.valid) and (s.flags & MF_SPRING)
	and ((mo.z + mo.height*2) >= s.z) and (mo.z <= (s.z + s.height/2))
		if not (mobjinfo[s.type].damage) then return end -- only diagonal springs should have damage set
		if not ((mobjinfo[s.type].damage/FRACUNIT) >= 26) then return end	-- blue springs destroyed us all

		local speed = (mobjinfo[s.type].damage/FRACUNIT)*3/2
		mo.player.speed = speed*mo.scale
	end
end, MT_PLAYER)

-- This is for Zone Battle's item boxes, to make them give zone energy instead of items.
addHook("MobjMoveCollide", function(mo, s)
	if not (btlzone and zoneon) return end
	if s and s.valid and s.type == MT_RANDOMITEM and mo.z + mo.height*2 >= s.z and mo.z <= s.z + s.height and s.health

		if mo.zboxcoords	-- coordinates of the last box hit:
			-- check for abs z dist and dist between xy
			if R_PointToDist2(s.x, s.y, mo.zboxcoords[1], mo.zboxcoords[2]) < 1024*s.scale
			and abs(s.z - mo.zboxcoords[3]) < 256*s.scale	-- we can assume s has the mapobjectscale :shitsfree:
				mo.zboxcoords = {s.x, s.y, s.z}	-- this doesn't mean you get the right to cheat anyway
				return false	-- nnnope!
			end
		end

		mo.zboxcoords = {s.x, s.y, s.z}	-- we will perform distance checks from this new box now.

		P_KillMobj(s, mo, mo)	-- kill box
		mo.player.kartstuff[k_itemroulette] = 0	-- no.

		z_fillzonegauge(mo)
		return true
	end
end, MT_PLAYER)


-- prevent collisions between players
local function noplayercollide(mo,mo2)

	if not zoneon or G_BattleGametype() return end

	if mo2 and mo2.valid and mo2.player
		return false
	end
end
addHook("MobjCollide", noplayercollide, MT_PLAYER)
addHook("MobjMoveCollide", noplayercollide, MT_PLAYER)
freeslot("sfx_sbrk")

-- global func to do zone damage if you want to use it or w/e
rawset(_G, "z_damage", function(p, d, f)
	local mo = p.mo
	if not p or not mo or not d return end

	if mo.zspinoutdeath return end
	if not mo.health return end
	if p.exiting return end
	if (p.kzonespeed or 0)/FRACUNIT >= dynzone and btlzone return end

	local targethp = p.zhp-(mo.zhploss+d)

	if mo.zshieldinvinc
		if p == displayplayer and not splitscreen
			P_StartQuake(40<<FRACBITS, 5)
		end
		return
	end

	if p.zhp == 0	-- no HP, we die.
		P_DamageMobj(mo)
		mo.zspinoutdeath = TICRATE
		p.kartstuff[k_spinouttimer] = TICRATE*8
		p.kartstuff[k_sneakertimer] = 0
		mo.zhudhake = TICRATE*2
		if netgame and p ~= displayplayer
			z_sendmsg(nil,"\x82"..p.name.." HAS BEEN ELIMINATED")
			S_StartSound(nil, sfx_zelmn)
		end
		S_StartSound(mo, sfx_khurt1+P_RandomRange(0, 1))
		if p == displayplayer and not splitscreen
			P_StartQuake(50<<FRACBITS, TICRATE*1 + 5)
		end

		if mo.skin:lower():find("wario") or not P_RandomRange(0, 31)
			S_StartSound(nil, sfx_zwaa)
		end

		return
	elseif targethp > 0	-- we have HP and will have some left after damage, shield us
		S_StartSound(mo, sfx_buzz2)
		K_DoInstashield(p)
	else				-- we have HP but won't have any after taking damage, show that by breaking our shield
		S_StartSound(mo, sfx_sbrk)
		for i = 1, 16 do
			local t = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
			t.scale = mo.scale
			t.color = 46
			t.destscale = 1
			t.scalespeed = mo.scale/12
			t.momx = P_RandomRange(-16,16)*mo.scale	+mo.momx
			t.momy = P_RandomRange(-16,16)*mo.scale	+mo.momy
			t.momz = P_RandomRange(-16,16)*mo.scale	+mo.momz
			t.tics = TICRATE*2
		end
	end

	mo.zhudshake = d

	mo.zhploss = $+d
	p.powers[pw_flashing] = $ or f or TICRATE
	mo.znoperfectlap = true

	-- Additions for Perfect Zone stuff
	mo.znoperfectzone = true
	p.perfectzonecombo = 0
	-- End of additions

	if p == displayplayer and not splitscreen
		P_StartQuake(40<<FRACBITS, 5)
	end
end)

local function z_handlehp(p)
	local mo = p.mo
	if mo.zspinoutdeath return end
	p.zhp = p.zhp == nil and starthp.value or p.zhp
	mo.zhploss = $ or 0
	if mo.zhploss

		-- No, so start draining their HP
		p.zhp = $-1
		mo.zhploss = $-1
		if not p.zhp
			mo.zhploss = 0 -- we can only die if we get hit at 0 HP.
		end
	end

	if mo.zhudshake
		mo.zhudshake = $-1
	end

	-- if we respawn, that means we've been sent off track or used the respawn command.
	-- either way, lose 10 HP

	-- unless we're playing zone battle and have respawned

	if p.kartstuff[k_respawn] == 47
		if btlzone and p.kzonerespawn == 2
			mo.zshieldboost = 100
			p.kzonerespawn = 1
		elseif p.kzonerespawn == 3
			p.kzonerespawn = nil
		else
			z_damage(p, 5, TICRATE*3)
			p.kzonerespawn = nil
		end
		if not p.kzonerespawn
			p.kzonerespawn = -1	-- default, used for normal 3 seconds shielding
		end
		mo.znoperfectlap = true	-- no perfect lap
	end

	-- we are spinning out.
	-- I don't care what caused you to spin out, you're gonna lose 5 HP as if you hit a wall
	if p.kartstuff[k_wipeoutslow]
	or p.kartstuff[k_spinouttimer]
		if not mo.zspinout
			mo.zspinout = true
			z_damage(p, 5)
		end
	else
		mo.zspinout = nil
	end

	if mo.zwallbounce
		mo.zwallbounce = $-1
	end

	-- normal wall damage
	if mo.eflags & MFE_JUSTBOUNCEDWALL
	or p.kartstuff[k_justbumped] == 5	// also do it if we're bouncing off objects
		mo.zwallbounce = $ and $+8 or 8	-- used to determine if we're in need of respawn due to gbj or w/e

		if mo.zwallbounce > 32	-- this is petty bad
			p.kzonerespawn = 3
			P_KillMobj(mo)
			mo.zwallbounce = nil
		end

		if not p.powers[pw_flashing]
			z_damage(p, 5)
		end
	end

	-- low shield beeps:
	if p.zhp <= starthp.value/20
		if leveltime%10 == 0
			S_StartSound(nil, sfx_ding, p)
		end
	elseif p.zhp <= starthp.value/5
		if leveltime%20 == 0
			S_StartSound(nil, sfx_ding, p)
		end
	end
end

local function z_zoneupdate(p)
	local mo = p.mo
	if mo.zspinoutdeath return end

		-- shield invulnerability
	if mo.zshieldinvinc
		mo.zshieldinvinc = $-1
		if mo.zshieldinvinc < 35
		and p.zzoneshield and p.zzoneshield.valid

			if leveltime%2
				p.zzoneshield.flags2 = $|MF2_DONTDRAW
			else
				p.zzoneshield.flags2 = $ & ~MF2_DONTDRAW
			end
		end

	else
		-- remove shield if it exists
		if p.zzoneshield and p.zzoneshield.valid
			P_RemoveMobj(p.zzoneshield)
		end
		p.zzoneshield = nil
	end

	local oldbrake = (P_IsObjectOnGround(mo) and p.cmd.buttons & BT_BRAKE and p.zhp < 2)

	if not p.kartstuff[k_respawn]	-- don't increment zone or score while we respawn
	and not oldbrake
	and mo.health
		p.zscore = $+1
		p.kzonespeed = $+zoneincrement / (btlzone and 3 or 1)
		-- ^ in zone battle, build zones much slower
		-- this is to pad out the games a bit more since high speeds can't be handled well with multiplayer input lag

		-- if you're drifting for long enough, charge zones faster!
		if p.kartstuff[k_driftcharge] > 1

			p.kartstuff[k_driftcharge] = 2	-- don't f*cking need it

			mo.z_drift = $ and $+1 or 1

			if mo.z_drift > 16

				p.kzonespeed = $ + zoneincrement/2 / (btlzone and 3 or 1)
				p.zscore = $+1

				if leveltime%2
					mo.colorized = true
					mo.z_decolorize = true
				end
			end
		else
			mo.z_drift = 0
		end
		if mo.z_decolorize and not mo.z_drift
			mo.colorized = false
			mo.z_decolorize = nil
		end

		-- check for floor boost sec special
		if P_PlayerTouchingSectorSpecial(p, 4, 6)	-- sneaker pannel spc
			if not mo.zpadtouch

				if not mo.zpadspeed or mo.zpadspeed < FRACUNIT/4	-- don't stack ridiculous speed pads
					p.kzonespeed = $ + FRACUNIT/2 / (btlzone and 2 or 1)
					p.zscore = $ + 50
				end

				mo.zpadspeed = FRACUNIT*3/4	-- speed pads give a boost. there's some smoothing going on though, see z_autothrust
				mo.zpadtouch = true
			end
		else
			mo.zpadtouch = nil	-- not touching, reset this
		end
	end

	if oldbrake
		p.zscore = max(0, $-10)	-- penalize braking
	end

	p.kartstuff[k_lapanimation] = 0	-- special case handled here, prevent lap anims from showing up.
	if mo.zoneupdate
		mo.zoneupdate = $-1
	end

	p.kartstuff[k_itemroulette] = 0
	p.kartstuff[k_itemtype] = 0

	-- Additions for Perfect Zone stuff
	if p.currentzone < p.kzonespeed / FRACUNIT -- We moved to a new Zone

		if not mo.zzoneboost	-- we are boosting through zones in zone battles, don't show all of this:

			-- play sfx and spawn some neat effect to show that we're going slightly faster
			local boost = P_SpawnMobj(mo.x, mo.y, mo.z, MT_BOOSTFLAME)
			boost.target = mo
			boost.tracer = mo
			boost.fuse = 7
			S_StartSound(nil, sfx_s23c, p)

			if zonesfx[p.currentzone+1]
				S_StartSound(nil, _G[zonesfx[p.currentzone+1]], p)
			else
				S_StartSound(nil, sfx_ptally, p)
			end

		end
		mo.zoneupdate = TICRATE	-- used to flash the "ZONE" when you get to a new one

		if not mo.zzoneboost

			if not mo.znoperfectzone
				p.perfectzones = $ + 1
				p.perfectzonecombo = $ + 1
				if p.perfectzonecombo > p.bestzonecombo then p.bestzonecombo = p.perfectzonecombo end
				p.zscore = $ + 50 + (50*p.perfectzonecombo)
				-- You get 100 points per perfect zone, plus an extra 50 for each zone in a combo
				if p.perfectzonecombo == 1 or btlzone	-- combos are useless in zone battle
					z_sendmsg(p, "\x82".."PERFECT ZONE (+2 SHIELD)")
				else
					z_sendmsg(p, "\x82".."PERFECT ZONE x"..p.perfectzonecombo.." (+2 SHIELD)")
				end

				p.zhp = min(starthp.value, $+2)
			end
		end
		mo.znoperfectzone = nil -- Reset perfect Zone
		p.currentzone = $ + 1
	end

	if p.laps > 0	-- if we have completed a lap...
		p.zlaps = $ or 0

		mo.zboxcoords = nil	-- reset itembox stuff
		p.zlaps = $+1

		if btlzone
			z_fillzonegauge(mo)
			z_sendmsg(p, "\x82".."ZONE GAUGE FILLED")
		end

		-- the problem with rally races is that the circuit doesn't loop, so let's send the player back to the start when they loop it once
		if mapheaderinfo[gamemap].levelflags & LF_SECTIONRACE
		and p.zlaps%(mapheaderinfo[gamemap].numlaps or 1) == 0
			-- reset position of players back to the position saved in z_init
			P_TeleportMove(mo, p.zx, p.zy, p.zz + mapheaderinfo[gamemap].mobj_scale*256)
			mo.momx, mo.momy, mo.momz = 0, 0, 0
			mo.angle = p.za
			p.kartstuff[k_respawn] = 46
			p.powers[pw_flashing] = TICRATE*3	-- small invuln for adapting to your new teleport
			z_sendmsg(p, "\x82".."COURSE RESET")
		end

		if not mo.znoperfectlap	-- we haven't hit anything, we get a perfect lap
		and not btlzone
			p.zhp = min(starthp.value, $+20)
			p.zscore = $+1000	-- perfect lap score
			z_sendmsg(p, "\x82".."PERFECT LAP")
			z_sendmsg(p, "\x82".."+20 SHIELD")
			S_StartSound(nil, sfx_yeeeah, p)
			p.perfectlaps = $ + 1
		end

		p.zscore = $+100	-- lap score
		mo.znoperfectlap = nil	-- reset perfect lap
		p.laps = 0	-- it doesn't count.
	end

end

local function z_autothrust(p)
	local mo = p.mo
	if mo.zspinoutdeath return end

	p.kartspeed = 8
	p.kartweight = 2	-- everyone gets sonic's stats for balance

	local speed = p.kzonespeed

	if btlzone	-- in zone battle, we will cut the speed in half. This is to make sure that this is fun to play for everyone
		speed = $/2
	end

	-- speed pad smoothing:
	if mo.zpadspeed
		mo.zpadsmoothspeed = $ or 0
		if mo.zpadsmoothspeed < mo.zpadspeed
			mo.zpadsmoothspeed = min(mo.zpadspeed, $ + FRACUNIT/10)
		else
			mo.zpadspeed = max(0, $ - FRACUNIT/128)
			mo.zpadsmoothspeed = mo.zpadspeed
		end

		-- particle effect because that's cool
		if leveltime%4 == 0 and mo.zpadspeed > FRACUNIT/4

			local facean = R_PointToAngle2(0, 0, mo.momx, mo.momy)
			local x = mo.x + FixedMul(mo.scale, 128*cos(facean))
			local y = mo.y + FixedMul(mo.scale, 128*sin(facean))

			local b = P_SpawnMobj(x, y, mo.z, MT_ZONEBOOSTINDICATOR)
			b.color = mo.color
			b.scale = mo.scale
			b.destscale = mo.scale*2
			b.scalespeed = mo.scale/3
			b.target = mo
			b.angle = mo.angle + ANG1*90
			K_FlipFromObject(b, mo)
		end
	end


	p.kartstuff[k_boostpower] = FRACUNIT*3/4 + speed/64 + (mo.zpadsmoothspeed or 0)
	mo.tspeed = FixedHypot(mo.momx, mo.momy)
	p.maxspeed = max($, mo.tspeed)	-- register highest speed for score

	if p.kartstuff[k_sneakertimer]
	and not p.kartstuff[k_wipeoutslow]
	and not p.kartstuff[k_spinouttimer]
		p.kartstuff[k_sneakertimer] = max($, 5)
	end
end


-- when playercmd is ready, use this method of killing the brakes instead:
addHook("PlayerCmd", function(p, cmd)
	if not zoneon return end
	if cmd.buttons & BT_BRAKE and p.zhp < 2 and P_IsObjectOnGround(p.mo)
		cmd.forwardmove = 50<<FRACBITS>>16
		cmd.buttons = $ & ~BT_BRAKE
	end
end)


-- handle brakes hp loss

freeslot("S_WZ_AIRBRAKE")	-- dummy state for air braking
states[S_WZ_AIRBRAKE] = {SPR_NULL, A, 2, nil, 0, 0, S_PARTICLE}

local function z_brakehandle(p)
	local mo = p.mo

	mo.z_brakepenalty = $ or 0
	mo.z_killbrake = false	-- see z_autothrust()
	if p.cmd.buttons & BT_BRAKE
		if P_IsObjectOnGround(mo)	-- if on ground, we still have control and deserve to lose HP for using brakes

			-- braking is violent, show sparks and smoke coming out of the tires!
			if not p.kartstuff[k_brakedrift]
				if not (leveltime%4)
					S_StartSound(mo, sfx_cdfm17)
					local side = P_RandomRange(0, 1)
					local sang = side and ANG1*70 or ANG1*-70
					local sx = mo.x + mo.momx + FixedMul(mo.radius, cos(mo.angle + ANG1*180 + sang))
					local sy = mo.y + mo.momy + FixedMul(mo.radius, sin(mo.angle + ANG1*180 + sang))
					local sparks = P_SpawnMobj(sx, sy, mo.z, MT_THOK)
					sparks.state = S_BRAKEDRIFT
					sparks.color = 15
					sparks.angle = mo.angle + 180*ANG1 + (P_RandomRange(20, 80)*ANG1*(side and -1 or 1))
					sparks.momx = mo.momx
					sparks.momy = mo.momy
					sparks.momz = P_MobjFlip(mo)*P_RandomRange(2, 6)*mo.scale
					sparks.flags2 = mo.flags2 & MF2_OBJECTFLIP
					sparks.target = mo
					sparks.scale = mo.scale
					sparks.fuse = TICRATE/4

					for i = 1, 5
						local smoke = P_SpawnMobj(sx, sy, mo.z + i*mo.scale*4, MT_SMOKE)
						smoke.scale = mo.scale/4
						smoke.destscale = mo.scale
						smoke.momx = mo.momx + P_RandomRange(-1, 1)*mo.scale
						smoke.momy = mo.momy + P_RandomRange(-1, 1)*mo.scale
						smoke.momz = mo.momz*3/4 + P_RandomRange(2, 5)*mo.scale
						smoke.scalespeed = mo.scale/12
					end
				end
			end

			-- don't allow braking at 1 HP, so that we can take an extra hit, and avoid any continuous cheese
			local minhp = 1
			if p.zhp > minhp	-- can't brake at 1/0hp. this is to avoid cheeses

				mo.z_brakepenalty = $+1

				if mo.z_brakepenalty > 4
					p.zhp = max(minhp, $-1)
					mo.z_brakepenalty = 0
				end
			else
				mo.z_killbrake = true
			end
			-- End of alterations

		else	-- otherwise, in mid air, allow free brakes
			local speed = FixedHypot(mo.momx, mo.momy) >> FRACBITS
			if speed > max(10, p.kzonespeed/FRACUNIT/4)
				local an = R_PointToAngle2(0, 0, mo.momx, mo.momy)
				P_Thrust(mo, an, -FRACUNIT*3/2)	-- slowly, SLOWLY thrust backwards

				for i = 1, 8
				local sx = mo.x + FixedMul(mo.radius*4, cos(an))
				local sy = mo.y + FixedMul(mo.radius*4, sin(an))

				local side = P_RandomRange(0, 1)

				local airbrake = P_SpawnMobj(sx, sy, mo.z+mo.height/3, MT_PARTICLE)
				airbrake.state = S_WZ_AIRBRAKE
				airbrake.scale = mo.scale*3
				airbrake.destscale = 1
				airbrake.scalespeed = mo.scale/12
				airbrake.angle = an + ANG1*180 + (P_RandomRange(90, 70)*ANG1*(side and -1 or 1))
				P_InstaThrust(airbrake, airbrake.angle, P_RandomRange(4, 16)*mo.scale)
				airbrake.momz = P_RandomRange(-30, 30)*mo.scale
				airbrake.fuse = 12
				airbrake.frame = A|TR_TRANS70
				end

				if not mo.z_brakesound
					S_StartSound(mo, sfx_ngskid)
					mo.z_brakesound = true
				end
			end
		end
	else
		mo.z_brakepenalty = 0
		mo.z_brakesound = nil
	end
end

-- speed fix
addHook("MobjThinker", function(mo)
	if not zoneon or G_BattleGametype() return end
	if leveltime < starttime return end
	mo.player.speed = 32*FRACUNIT
end, MT_PLAYER)

local function z_init(p)
	local mo = p.mo
	if not mo return end
	if not p.z_init
		mo.z_brakepenalty = 0
		p.zhp = starthp.value
		mo.zhploss = 0
		mo.zhudshake = 0
		p.zscore = 0
		p.zlaps	= 0	-- used for section races
		p.kzonespeed = FRACUNIT*startzone.value

		p.currentzone = startzone.value
		p.perfectzones = 0
		p.perfectzonecombo = 0
		p.bestzonecombo = 0
		p.perfectlaps = 0

		-- battle mode variables
		p.zbattleboost = 0		-- boost ranges from 0 to FRACUNIT*6
		p.zbattleshield = 0		-- shield ranges from 0 to 100

		p.maxspeed = 0
		p.zx = mo.x
		p.zy = mo.y
		p.zz = mo.z
		p.za = mo.angle -- save coords for respawning in segment races

		-- haha big funni hacker, i see you coming:
		p.zmodified = startzone.value > 1 or starthp.value > 100

		p.z_init = true
	end
end

-- synch zoneon and btlzone
addHook("NetVars", function(net)
	zoneon = net(zoneon)
	btlzone = net(btlzone)
	dynzone = net(dynzone)
	battlefinish = net(battlefinish)
end)

-- count players in the game
local function z_countplayers(f)
	local count = 0
	for p in players.iterate do
		if not p.mo continue end
		if f
			if f(p)
				count = $+1
			end
		else
			count = $+1
		end
	end
	return count
end

-- ZONE BARRIERS:

function A_ZoneBarrier(a)

	if not a or not a.valid return end

	if not a.savecolor
		a.savecolor = a.color
	end

	if leveltime%10 < 5
		a.color = ColorOpposite(a.savecolor)
	else
		a.color = a.savecolor
	end
end

freeslot("MT_ZONEBARRIER")
freeslot("S_ZONEBARRIER")
freeslot("SPR_ZBRR")
mobjinfo[MT_ZONEBARRIER] = {
	mapthingnum = -1,
	spawnstate = S_ZONEBARRIER,
	spawnhealth = 1000,
	radius = 32<<FRACBITS,
	height = 64<<FRACBITS,
	flags = MF_NOGRAVITY
}
states[S_ZONEBARRIER] = {SPR_ZBRR, A|FF_FULLBRIGHT|FF_ANIMATE|TR_TRANS20, 1, A_ZoneBarrier, J, 1, S_ZONEBARRIER} -- was PAPERSPRITE / TRANS40 but fuck that

-- ZONE SHIELD
for i = 1, 6
	freeslot("S_ZONESHIELD"..i)
end

states[S_ZONESHIELD1] = {SPR_ISTA, A|FF_FULLBRIGHT|TR_TRANS30, 3, nil, 0, 0, S_ZONESHIELD2}
states[S_ZONESHIELD2] = {SPR_ISTA, B|FF_FULLBRIGHT|TR_TRANS30, 3, nil, 0, 0, S_ZONESHIELD3}
states[S_ZONESHIELD3] = {SPR_ISTA, C|FF_FULLBRIGHT|TR_TRANS30, 3, nil, 0, 0, S_ZONESHIELD4}
states[S_ZONESHIELD4] = {SPR_ISTA, D|FF_FULLBRIGHT|TR_TRANS30, 1, nil, 0, 0, S_ZONESHIELD5}
states[S_ZONESHIELD5] = {SPR_ISTA, E|FF_FULLBRIGHT|TR_TRANS30, 1, nil, 0, 0, S_ZONESHIELD6}
states[S_ZONESHIELD6] = {SPR_ISTA, F|FF_FULLBRIGHT|TR_TRANS30, 1, nil, 0, 0, S_ZONESHIELD4}

-- ZONE BOOSTER
-- slingshot level indicators
freeslot("MT_ZONEBOOSTINDICATOR", "SPR_SLNS")
for i = 1,3
	freeslot("S_ZONEBOOSTINDICATOR"..i)
end
for i = 1, 3
	local tr = {TR_TRANS60, TR_TRANS70, TR_TRANS80}
	states[S_ZONEBOOSTINDICATOR1 + (i-1)] = {SPR_SLNS, (i-1)|tr[i]|FF_PAPERSPRITE|FF_FULLBRIGHT, 2, nil, 0, 0, i < 3 and (S_ZONEBOOSTINDICATOR1 +i) or S_NULL}
end

mobjinfo[MT_ZONEBOOSTINDICATOR] = {
	spawnstate = S_ZONEBOOSTINDICATOR1,	-- must be modified on spawn, our script takes care of it (...)
	deathstate = S_NULL,
	spawnhealth = 1000,
	radius = 65*FRACUNIT,
	height = 132*FRACUNIT,
	dispoffset = 2,
	flags = MF_NOGRAVITY|MF_DONTENCOREMAP|MF_NOCLIPHEIGHT
}

-- zone barrier hit thinker:
addHook("MobjCollide", function(mo, b)
	if not zoneon return end
	if b and b.valid and b.type == MT_ZONEBARRIER and b and b.valid and mo.z + mo.height*2 >= b.z and mo.z <= b.z + b.height/2
		-- we hit a barrier, what do we do?
		-- case 1: this is our new fresh barrier, or we have flashtics, or are protected
		if mo == b.target and leveltime-b.threshold < TICRATE or mo.player.powers[pw_flashing] or mo.zshieldinvinc or mo.player.exiting
			return false
		end	-- we never collided with it.


		if hitfeed and b and b.valid
			--HF_SendHitMessage(b.target.player, mo.player, "ZONEHF", b.savecolor)
			local sourcep

			if b.target and b.target.valid and b.target.player	-- player still exists
				sourcep = b.target.player
			else	-- player has died, let's check with savename

				for p in players.iterate
					if b.savename == p.name	-- this will do
					and p.mo and p.mo.valid
						sourcep = p
						break
					end
				end
			end

			if not sourcep return end	-- i give up
			local source

			if sourcep ~= mo.player

				source = {
					{sourcep.name, sourcep.mo.color},
					{skins[sourcep.mo.skin].facemmap, sourcep.mo.color},
					sourcep,
				}
			end	-- Leave source nil if we are our own murderer.

			local victim = {
				{mo.player.name, mo.color},
				{skins[mo.skin].facemmap, mo.color},
				mo.player,
			}

			HF_SendHitMessage(source, victim, "ZONEHF", b.savecolor)
		end

		-- case 3: this is a barrier, and we have no shield, take the damage.
		if mo == b.target
			P_DamageMobj(mo, nil, nil)		-- usually, we can't damage ourselves.
		else
			P_DamageMobj(mo, b, b.target)
		end
		P_InstaThrust(mo, R_PointToAngle2(0, 0, mo.momx, mo.momy), FixedHypot(mo.momx, mo.momy)/4)
		mo.player.kartstuff[k_sneakertimer] = 0	-- cut our momentum, this gives us time to readjust ourselves.
		z_damage(mo.player, 20)
		mo.player.zbattleboost = $/2
		mo.player.zbattleshield = $/2

		-- break barrier:
		for i = 1, 16 do
			local t = P_SpawnMobj(b.x, b.y, b.z, MT_THOK)
			t.scale = mo.scale
			t.color = 44
			t.destscale = 1
			t.scalespeed = mo.scale/12
			t.momx = P_RandomRange(-16,16)*b.scale	+mo.momx
			t.momy = P_RandomRange(-16,16)*b.scale	+mo.momy
			t.momz = P_RandomRange(-16,16)*b.scale	+mo.momz
			t.tics = TICRATE*2
		end
		P_RemoveMobj(b)
		return false
	end
end, MT_PLAYER)

-- spawn shield for zone battle and also as a normal item, eventually?
local function z_doshield(p, timer)
	local mo = p.mo
	if not mo return end
	mo.zshieldinvinc = timer

	if p.zzoneshield and p.zzoneshield.valid
		P_RemoveMobj(p.zzoneshield)
	end

	local b = P_SpawnMobj(mo.x, mo.y, mo.z, MT_OVERLAY)
	b.state = S_ZONESHIELD1
	b.target = mo
	p.zzoneshield = b
end

-- zone battle shit goes here
local function z_handlebattlestuff(p)
	local mo = p.mo
	if p.exiting return end

	-- our zone is above the target zone, we won!
	if p.kzonespeed >= dynzone<<FRACBITS
		p.kzonespeed = dynzone<<FRACBITS
		mo.kzonefinish = battlefinish*(dynzone<<FRACBITS)	-- make sure we don't go above, this would screw up the rankings
		mo.kzonepos = 17 - battlefinish
		p.kartstuff[k_position] = mo.kzonepos
		battlefinish = $-1
		P_DoPlayerExit(p)
		S_StopSound(mo)
		return
	end

	-- check if we have more than 15% of zone gauge filled.
	if p.zbattleshield > battlemaxshield*2/13	-- we can assume that %zone == %shield on the gauge
	and not mo.zzoneboost
	and not mo.zshieldboost
	and mo.health
	and not mo.zspinoutdeath

		-- from here on out, we can use zone boost or zone absorb
		-- ready inputs with c1 and c2
		mo.zc1in = $ or 0
		mo.zc2in = $ or 0

		mo.zc1in = (p.cmd.buttons & BT_ATTACK and not (p.cmd.buttons & BT_BACKWARD)) and $+1 or 0
		mo.zc2in = (p.cmd.buttons & BT_CUSTOM1 or (p.cmd.buttons & BT_ATTACK and p.cmd.buttons & BT_BACKWARD)) and $+1 or 0

		-- zone boost:
		if mo.zc1in == 1
			mo.zzoneboost = p.zbattleboost
			p.zbattleboost, p.zbattleshield = 0, 0

			-- this gets us the neat camera effect and boost as we skim through zones.
			p.kartstuff[k_sneakertimer] = 0
			p.kartstuff[k_floorboost] = 0
			p.kartstuff[k_speedboost] = 0
			K_DoSneaker(p)
			K_PlayAttackTaunt(mo)

			-- spawn the zone barrier:
			local b = P_SpawnMobj(mo.x, mo.y, mo.z, MT_ZONEBARRIER)
			b.angle = R_PointToAngle2(0, 0, mo.momx, mo.momy) + 90*ANG1	-- it's a papersprite, rotate it
			b.target = mo
			b.scale = mo.scale
			b.color = mo.color
			b.threshold = leveltime	-- while threshold < leveltime+TICRATE, b.target cannot collide with b
			b.savename = p.name	-- hack

			-- calculate the zone we're gonna end up at and play its sound instantly; this is to avoid getting earraped
			local newzone = p.kzonespeed/FRACUNIT + mo.zzoneboost/FRACUNIT
			local perfectzones = 0
			local nosfx = 0

			while newzone
				if newzone <= p.kzonespeed/FRACUNIT
					break
				end

				if not mo.znoperfectzone
					perfectzones = $+1
				else
					mo.znoperfectzone = nil
				end

				if zonesfx[newzone] and nosfx
					S_StartSound(nil, _G[zonesfx[newzone]], p)
					nosfx = true
				end
				newzone = $-1
			end

			-- give shield back accordingly with perfect zones
			if perfectzones
				p.zhp = $ + 2*perfectzones
				z_sendmsg(p, "\x82".."PERFECT ZONE x"..perfectzones.." (+"..(2*perfectzones).." SHIELD)")
			end

			-- finally, apply a boost!
			mo.zpadspeed = FRACUNIT*3/4

		-- barrier absorb
		elseif mo.zc2in == 1
			mo.zshieldboost = p.zbattleshield
			z_doshield(p, p.zbattleshield*3)
			p.zbattleboost, p.zbattleshield = 0, 0
			K_PlayBoostTaunt(mo)
			S_StartSound(mo, sfx_buzz3)
		end
	end

	if not mo.health return end
	if mo.zspinoutdeath return end

	-- handle zone boost:
	if mo.zzoneboost
		p.kzonespeed = $+ min(mo.zzoneboost, FRACUNIT/2)
		mo.zzoneboost = max($-FRACUNIT/2, 0)
	end
	-- handle shield boost:
	if mo.zshieldboost
		p.zhp = $+ min(mo.zshieldboost, 5)
		p.zhp = min($, starthp.value)
		mo.zshieldboost = max($-5, 0)
	end
end

-- some variables for zone battle timeout
local zonetimeout = 0		-- flash "ZONE TIMEOUT" text
local timeoutset = false	-- make sure we don't do the above everytime
local zonetargetflash = 0	-- flashes the "TARGET" to show that it's doing something
-- note: nobody cares if those are actually synched

-- set values on map load
addHook("MapLoad", do

	battlefinish = 16
	zonetimeout = 0
	timeoutset = false
	zonetargetflash = 0

	-- reenable hud items on map load
	hud.enable("time")
	hud.enable("minimap")
	hud.enable("speedometer")
	hud.enable("minirankings")
	hud.enable("gametypeinfo")
	hud.enable("position")
	hud.enable("item")
	hud.enable("freeplay")

	-- 1: if we have the cvar enabled, turn on wipezones, off otherwise
	zoneon = haszone.value
	btlzone = (forcebattle.value) or (hasbattle.value and z_countplayers() > 1)
	dynzone = zonetarget.value	-- dynamc target for zone battle

	if not zoneon or G_BattleGametype()
		zoneon = false
		btlzone = false
		return
	end

	-- 2: reset player vars:
	for p in players.iterate do
		p.z_init = nil
		if p.mo and not p.spectator
			z_init(p)
		end
	end

	-- 3: if we have wipezones, remove item boxes
	-- Do not do it for Zone Battle however, since they serve a purpose there.

	if btlzone return end

	for m in thinkers.iterate("mobj")
		if m and m.valid and m.type == MT_RANDOMITEM
			m.state = S_INVISIBLE
			m.tics = 1
			m.flags = MF_NOBLOCKMAP
		end
	end
end)

-- display results after a time over
local function z_results(p)
	local mo = p.mo
	mo.zresults = $ or 0
	mo.zresults = $+1

	if mo.zresults > TICRATE*6
	and p.cmd.buttons & BT_CUSTOM1
	--or mo.zresults > TICRATE*20
		mo.z_exit = true
	end

	if mo.zresults > TICRATE*6
		local c = z_countplayers()
		local e = z_countplayers(function(p) return p.mo.z_exit end)

		if e >= c
		and not server.exiting
			z_countplayers(function(p) p.exiting = TICRATE end)
		end
	end
end

local function z_deathseq(p)
	local mo = p.mo
	mo.zhudshake = 69	-- force hud to shake so that it looks cool
	mo.zspinoutdeath = $-1

	if btlzone

		-- in zone battle,
		-- explode in style, lose 5 zones, and respawn asap

		if not mo.zspinoutdeath
			K_SpawnMineExplosion(mo)
			P_KillMobj(mo, nil, nil)
			mo.zspinoutdeath = nil
			p.zbattleshield = 0
			p.zbattleboost = 0
			p.kzonespeed = max(FRACUNIT, $-(5<<FRACBITS))
			p.kzonerespawn = 2	-- if kzonerespawn == 2: give 100 HP back on respawn, if 1, set invincibility when you hit the floor
		end
		return
	end

	if not mo.zspinoutdeath
		z_blowupplayer(p)
	end
end

-- some visual stuff
local function z_visualquirks(p)
	local mo = p.mo
	-- shake the hud and camera if you hit the ground too hard. It gives a sense of speed and weight.
	if mo.zprevmomz ~= mo.momz

		if mo.eflags & MFE_JUSTHITFLOOR
		and not P_CheckDeathPitCollide(mo)

			if p.kzonerespawn == 1 or p.kzonerespawn == -1	-- set flashtics
				z_doshield(p, TICRATE* (p.kzonerespawn == 1 and 8 or 3))
				p.kzonerespawn = nil
			end

			if P_MobjFlip(mo)*mo.zprevmomz < mo.scale*-8
				mo.zhudshake = 8
				S_StartSound(mo, sfx_doord2)
				if p == displayplayer
					P_StartQuake(40<<FRACBITS, 6)
				end
			else
				mo.zhudshake = 2
				S_StartSound(mo, sfx_cybdth)
				if p == displayplayer
					P_StartQuake(10<<FRACBITS, 4)
				end
			end
		end

		mo.zprevmomz = mo.momz
	end
end

-- make things easier on our end
local function z_battledynamictarget()

	-- if all players have exited, or are dead, don't continue
	if z_countplayers() == z_countplayers(function(p) return (p.exiting or p.pflags & PF_TIMEOVER) end)
		return
	end

	local startcountdown = TICRATE*60*4 +starttime
	/*
		By default, we will start decrementing zones if no one has reached the target in 4 minutes.
	*/

	-- check if a player has finished:
	for p in players.iterate do
		if p and p.valid and p.mo and p.mo.kzonepos and p.mo.kzonepos == 1 and p.realtime < startcountdown
			startcountdown = p.realtime	-- take first's finish time as a base if it's smaller than 4 mins
		end
	end

	if leveltime > startcountdown

		zonetimeout = $ and $-1 or 0
		zonetargetflash = $ and $-1 or 0

		if (leveltime-startcountdown) % (TICRATE*5) == 0
			S_StartSound(nil, sfx_ztime)
			zonetargetflash = TICRATE
			if not timeoutset
				zonetimeout = TICRATE*3
				timeoutset = true
			end
			dynzone = $-1	-- eventually everyone will win!
		end
	end
end

local hitfeedload

addHook("ThinkFrame", do
	if not zoneon or G_BattleGametype() return end

	if not hitfeedload	-- load hitfeed zone coords
		if hitfeed
			HF_OverrideLocation(9)
			hitfeedload = true
		end
	end

	for p in players.iterate do
		if not p.mo or p.spectator continue end
		if p.pflags & PF_TIMEOVER
		and not btlzone
			z_results(p)
			continue
		elseif p.mo.zspinoutdeath
			z_deathseq(p)
			continue
		end

		if p.exiting
			p.mo.zhudshake = 0
			continue
		end	-- don't do anything if we're exiting the map

		if leveltime > starttime
		and P_IsObjectOnGround(p.mo)
		and not p.kartstuff[k_sneakertimer]
		and not p.kartstuff[k_wipeoutslow]
		and not p.kartstuff[k_spinouttimer]
			K_DoSneaker(p)
		end

		if not p.kzonespeed
			p.z_init = nil
		end
		z_init(p)
		if leveltime < starttime return end	-- no need for zones if we can't try em

		z_zoneupdate(p)
		z_autothrust(p)
		z_brakehandle(p)
		z_handlehp(p)
		z_visualquirks(p)
		if btlzone
			z_handlebattlestuff(p)
		end
	end
	if leveltime >= starttime and btlzone
		z_battledynamictarget()
	end
end)



-- HUD RENDERING GOES HERE


-- draws a small circle (actually 2 rectangles but w/e lol)
-- this is used for the zone time display
local function hud_drawsmallcircle(v, x, y, c)
	v.drawFill(x, y+1, 4, 2, c)
	v.drawFill(x+1, y, 2, 4, c)
end

local function hud_zdispzone(v, p, c)
	if not p.mo return end
	-- we save this space for the rankings and target zone

	local rx, ry = 0, 0
	if p.mo.zhudshake and hudshake.value
		rx = N_RandomRange(-4, 4)
		ry = N_RandomRange(-4, 4)	-- shaker for damage
	end

	-- handle splitscreen:
	if splitscreen == 1
		local flagsets = {V_SNAPTOLEFT|V_SNAPTOTOP, V_SNAPTOLEFT}
		local coords = {{2, 12}, {2, 112}}
		local n = c.pnum
		if not n return end	-- ????

		local znum = p.kzonespeed and p.kzonespeed/FRACUNIT or 0
		local name = z_getzonename(znum)

		--if not ((p.mo.zoneupdate or p.mo.zoneupdate%10 < 5) or (zonetargetflash and zonetargetflash%10
		if not ((zonetargetflash and zonetargetflash%10 < 5) or (p.mo.zoneupdate and p.mo.zoneupdate%10 < 5))
			V_drawString(v, coords[n][1]+rx, coords[n][2]+ry, "ZONE "..z_getdec(p.kzonespeed)..(btlzone and "/"..dynzone..".0" or ""), "MKFNT", flagsets[n]|V_HUDTRANS)
		end
		if not btlzone	--draw score instead
			V_drawString(v, coords[n][1]+rx, coords[n][2]+12+ry, "SCORE "..p.zscore, "MKFNT", flagsets[n]|V_HUDTRANS, not front and "right" or nil)
		end
		return
	elseif splitscreen > 1
		local flagsets = {V_SNAPTOLEFT|V_SNAPTOTOP, V_SNAPTORIGHT|V_SNAPTOTOP, V_SNAPTOLEFT, V_SNAPTORIGHT}
		local coords = {{2, 12}, {318, 12}, {2, 112}, {318, 112}}
		local n = c.pnum
		local front = n%2
		if not n return end	-- ????

		if not ((zonetargetflash and zonetargetflash%10 < 5) or (p.mo.zoneupdate and p.mo.zoneupdate%10 < 5))
			V_drawString(v, coords[n][1]+rx, coords[n][2]+ry, "ZONE "..z_getdec(p.kzonespeed)..(btlzone and "/"..dynzone..".0" or ""), "MKFNT", flagsets[n]|V_HUDTRANS, not front and "right" or nil)
		end
		if not btlzone	--????
			V_drawString(v, coords[n][1]+rx, coords[n][2]+12+ry, "SCORE "..p.zscore, "MKFNT", flagsets[n]|V_HUDTRANS, not front and "right" or nil)
		end

		return
	end

	-- some may argue that speed would be part of the zone display since they're both related

	local spd = FixedDiv(FixedMul(p.mo.tspeed or 0, 142371), mapheaderinfo[gamemap].mobj_scale)/FRACUNIT

	V_drawString(v, 310-rx, 8-ry, "SPEED", "MKFNT", V_SNAPTORIGHT|V_SNAPTOTOP|V_HUDTRANS, "right", true)
	V_drawString(v, 310-rx, 24-ry, spd.." KM/H", "MKFNT", V_SNAPTORIGHT|V_SNAPTOTOP|V_HUDTRANS, "right", true)

	if btlzone return end	-- still draw the speedometer in zone battle

	-- note: x, y coordinates are the center of the "circle".
	local x, y = 320, 100
	local an = ANG1*195
	local rad = 300

	local znum = p.kzonespeed and p.kzonespeed/FRACUNIT or 0
	local name = z_getzonename(znum)

	-- draw zone progress at the top of the screen
	local cx = 12

	if leveltime > 105
	and p.kzonespeed ~= nil

		-- zone time calculator.
		-- these dots fill as you get closer to the next zone

		local progress = leveltime >= starttime and ((p.kzonespeed%FRACUNIT) *6 /FRACUNIT) +1 or 0
		-- *5/3 to allow the number to grow larger than 6. this is used below

		for i = 1, min(6, (leveltime-105)/4)

			if p.mo.zhudshake and hudshake.value
				rx = N_RandomRange(-4, 4)
				ry = N_RandomRange(-4, 4)	-- shaker for damage
			end

			local color = i > progress and 16 or 0

			if progress >= 6
			and p.kzonespeed%(zoneincrement*10) < zoneincrement*5
				color = 16
			end

			hud_drawsmallcircle(v, cx+rx, 3+ry, 31|V_SNAPTOTOP|V_SNAPTOLEFT)
			hud_drawsmallcircle(v, cx+rx, 2+ry, color|V_SNAPTOTOP|V_SNAPTOLEFT)
			cx = $+8
		end
	end

	if not p.mo.zoneupdate or p.mo.zoneupdate%10 < 5
		V_drawString(v, 10+rx, 8+ry, "ZONE", "MKFNT", V_SNAPTOLEFT|V_SNAPTOTOP|V_HUDTRANS, nil, true)
	end

	-- Start of addditions for limiting the shown Zones to just the current and next - Latius
	local nextzone = true -- Line Added
	for i = 0, 7
		local dx = x<<FRACBITS + rad*cos(an)
		local dy = y<<FRACBITS + rad*sin(an)

		if p.mo.zhudshake and hudshake.value
			rx = N_RandomRange(-4, 4)
			ry = N_RandomRange(-4, 4)	-- shaker for damage
		end

		if i and zonenames[znum+i] and nextzone -- Line added to
			name = zonenames[znum+i]
			nextzone = false -- Line added
		elseif i
			name = nil
		end

		V_drawString(v, dx>>FRACBITS +rx +2, dy >>FRACBITS +ry, tostring(znum +i), "MKFNT", V_SNAPTOLEFT|V_SNAPTOTOP|V_HUDTRANS, "center", true)
		if name
			V_drawString(v, dx>>FRACBITS + 6 + V_stringWidth(v, tostring(znum+i), "MKFNT", nil, true)/2 +rx, dy>>FRACBITS + 4 +ry, name:upper(), "MKFNT", V_SNAPTOLEFT|V_SNAPTOTOP|V_HUDTRANS)
		end
		--v.drawString(dx>>FRACBITS, dy>>FRACBITS, (p.p.kzonespeed and p.p.kzonespeed/FRACUNIT or 0) +i, V_SNAPTORIGHT)
		an = $-ANG1*4
	end
	-- End of additions
end

-- local var since this is really only purely visual
local client_boost = 0
local client_shield = 0

local function hud_zdispzonegauge(v, p, c)

	local rx, ry = 0, 0
	if p.mo.zhudshake and hudshake.value
		rx = N_RandomRange(-4, 4)
		ry = N_RandomRange(-4, 4)	-- shaker for damage
	end
	local time = leveltime - 105

	local boostcolors = {0, 24, 88, 103, 116, 123, 128, 168, 184, 190, 204, 231}
	local boostcond = (p.zbattleboost + p.kzonespeed > (dynzone<<FRACBITS)) and (leveltime%255) or 0
	local boosttowin = boostcond and boostcolors[N_RandomRange(1, #boostcolors)]

	-- handle splitscreen:
	if splitscreen == 1
		local flagsets = {V_SNAPTOLEFT, V_SNAPTOBOTTOM|V_SNAPTOLEFT}
		local coords = {{115, 98}, {115, 198}}
		local n = c.pnum
		local front = n%2
		if not n return end	-- ????

		-- bar OLs

		local basex = coords[n][1]
		local basey = coords[n][2]

		local minx = basex - 100
		local maxx = basex + 100

		v.drawFill(max(minx, basex - time*4) +rx, basey+ry, min(maxx-minx, time*8) , 1, flagsets[n])	-- bottom bar
		v.drawFill(basex-1+rx, max(basey-8, basey-time*4) +ry, 2, min(8, time*4), flagsets[n])	-- middle

		-- draw bars
		local blen = p.zbattleshield*100/(battlemaxshield)
		v.drawFill(basex+1, basey-6, blen, 5, (blen <= 15 and 128 or 185)|flagsets[n])
		blen = p.zbattleboost*100/(battlemaxboost)
		v.drawFill(basex-1-blen, basey-6, blen, 5, (boosttowin or (blen <= 15 and 128 or 185))|flagsets[n])

		-- draw values
		V_drawString(v, minx-8+rx, basey+ry - 14, z_getdec(p.zbattleboost), "MKFNT", flagsets[n]|V_HUDTRANS, nil, true)
		V_drawString(v, basex-2+rx, basey+ry - 5, "BOOST", "MKFNT", flagsets[n]|V_HUDTRANS, "right")

		V_drawString(v, maxx+8+rx, basey+ry - 14, tostring(p.zbattleshield)..".0", "MKFNT", flagsets[n]|V_HUDTRANS, "right", true)
		V_drawString(v, basex+2+rx, basey+ry - 5, "ABSORB", "MKFNT", flagsets[n]|V_HUDTRANS)
		return
	elseif splitscreen > 1
		local flagsets = {V_SNAPTOLEFT, V_SNAPTORIGHT, V_SNAPTOLEFT|V_SNAPTOBOTTOM, V_SNAPTORIGHT|V_SNAPTOBOTTOM}
		local coords = {{45, 98}, {275, 98}, {45, 198}, {275, 198}}
		local n = c.pnum
		local front = n%2
		if not n return end	-- ????

		-- bar OLs

		local basex = coords[n][1]
		local basey = coords[n][2]

		local minx = basex - 32
		local maxx = basex + 32

		v.drawFill(max(minx, basex - time) +rx, basey+ry, min(maxx-minx, time*2) , 1, flagsets[n])	-- bottom bar
		v.drawFill(basex-1+rx, max(basey-8, basey-time) +ry, 2, min(8, time), flagsets[n])	-- middle

		-- draw bars
		local blen = p.zbattleshield*32/(battlemaxshield)
		v.drawFill(basex+1, basey-4, blen, 3, (blen <= 4 and 128 or 185)|flagsets[n])
		blen = p.zbattleboost*32/(battlemaxboost)
		v.drawFill(basex-1-blen, basey-4, blen, 3, (boostowin or (blen <= 4 and 128 or 185))|flagsets[n])

		-- draw values
		V_drawString(v, minx+8+rx, basey+ry - 8, z_getdec(p.zbattleboost), "MKFNT", flagsets[n]|V_HUDTRANS)
		--v.drawString(basex-2+rx - v.stringWidth("BOOST", nil, "thin"), basey - 8, "BOOST", "MKFNT", flagsets[n]|V_HUDTRANS, "thin")

		V_drawString(v, maxx-8+rx, basey+ry - 8, tostring(p.zbattleshield)..".0", "MKFNT", flagsets[n]|V_HUDTRANS, "right")
		--v.drawString(basex+2+rx, basey+ry - 8, "SHIELD", "MKFNT", flagsets[n]|V_HUDTRANS, "thin")
		return
	end

	local bar_len_boost, bar_len_shield

	-- draw the gauge "outlines" at the bottom

	v.drawFill(max(50, 160 - time*4) +rx, 190+ry, min(220, time*8) , 1, V_SNAPTOBOTTOM)	-- bottom bar
	v.drawFill(158+rx, max(175, 190-time*4) +ry, 4, min(15, time*4), V_SNAPTOBOTTOM)	-- middle


	-- draw values
	-- boost

	-- update local len:
	if client_boost ~= p.zbattleboost
		client_boost = $+(p.zbattleboost-client_boost)/2
	end
	local blen = client_boost*100/(battlemaxboost)

	-- gauge
	v.drawFill(158 - blen+rx, 180+ry, blen, 8, (boosttowin or (blen <= 15 and 128 or 185))|V_SNAPTOBOTTOM)
	if p.zbattleboost	-- bar end
		for i = 0, 3
			v.drawFill(158-blen-1-i+rx, 180+ry +i*2, 1+i, 2, (boosttowin or (blen <= 15 and 128 or 185))|V_SNAPTOBOTTOM)
		end
	end

	--text
	V_drawString(v, 60+rx, 184+ry, "BOOST", "MKFNT", V_HUDTRANS|V_SNAPTOBOTTOM)
	V_drawString(v, 5+rx, 180+ry, z_getdec(client_boost), "MKFNT", V_HUDTRANS|V_SNAPTOBOTTOM, nil, true)

	-- absorb
	-- gauge

	-- update local len:
	if client_shield ~= p.zbattleshield
		client_shield = $+(p.zbattleshield-client_shield)/2
		if abs(client_shield - p.zbattleshield) < 4
			client_shield = p.zbattleshield
		end
	end
	blen = client_shield*100/(battlemaxshield)

	v.drawFill(162+rx, 180+ry, blen, 8, (blen <= 15 and 128 or 185)|V_SNAPTOBOTTOM)
	if p.zbattleshield	-- bar end
		for i = 0, 3
			v.drawFill(162+blen+rx, 180+ry +i*2, 1+i, 2, (blen <= 15 and 128 or 185)|V_SNAPTOBOTTOM)
		end
	end
	--text
	V_drawString(v, 260+rx, 184+ry, "ABSORB", "MKFNT", V_HUDTRANS|V_SNAPTOBOTTOM, "right")
	V_drawString(v, 316+rx, 180+ry, tostring(client_shield)..".0", "MKFNT", V_HUDTRANS|V_SNAPTOBOTTOM, "right", true)
end

local function hud_zdisphpandscore(v, p, c)

	if not p.mo return end
	if p.zhp == nil return end

	-- random stuff for taking damage:
	local rx, ry = 0, 0
	if p.mo.zhudshake and hudshake.value
		rx = N_RandomRange(-4, 4)
		ry = N_RandomRange(-4, 4)	-- shaker for damage
	end

	local lowhplevel = starthp.value / 5 -- Line added to get the Low HP Level dependant on the Max HP
	local redhp = p.zhp <= lowhplevel and leveltime%10 < 5
	local shieldcount = p.mo.zshieldinvinc and leveltime%10 < 5
	if p.zhp <= lowhplevel/4
		redhp = leveltime%5 < 2
	end


	-- HP, or something:
	local hstart = 40
	local hlen = 20
	local startlen = 20
	local hashp = p.zhp*hlen/starthp.value -- Altered to have it where the length is based on the Max HP
	-- End of additions

	local xmod = 0
	local xrev = false

	if leveltime < starttime
		startlen = (leveltime-105)/2
		startlen = max(0, $)
		startlen = min(20, $)
	end

	if splitscreen == 1		-- 2 players splitscreen
		-- draw the HP bar horizontally
		local flagsets = {V_SNAPTOLEFT|V_SNAPTOTOP, V_SNAPTOLEFT}
		local coords = {{26, 2}, {26, 102}}
		local n = c.pnum
		local front = n%2

		if not n return end	-- ????

		-- HP BAR:
		local basex = coords[n][1]
		local basey = coords[n][2]
		for i = 1, leveltime >= starttime and hlen*2 or startlen*2
			if p.mo.zhudshake and hudshake.value
				rx = N_RandomRange(-4, 4)
				ry = N_RandomRange(-4, 4)	-- shaker for damage
			end
			v.drawFill(basex + (i-1)*2 +rx, basey+ry, 1, 8, ((shieldcount and 185) or (redhp and i > hashp*2) and (128) or (i > hashp*2) and (16) or (215))|flagsets[n])
		end
		if p.mo.zshieldinvinc
			v.drawString(basex - 24, basey+ry, (p.mo.zshieldinvinc/35).."."..G_TicsToCentiseconds(p.mo.zshieldinvinc)/10, flagsets[n]|V_GREENMAP|V_HUDTRANS)
		else
			v.drawString(basex - 24, basey+ry, tostring(p.zhp), flagsets[n]|(redhp and V_REDMAP or V_BLUEMAP)|V_HUDTRANS)
		end

		if btlzone
			hud_zdispzonegauge(v, p, c)
		end

		return
	elseif splitscreen > 1	-- 3-4 player splitscreen
		-- draw the HP bar horizontally
		local flagsets = {V_SNAPTOLEFT|V_SNAPTOTOP, V_SNAPTORIGHT|V_SNAPTOTOP, V_SNAPTOLEFT, V_SNAPTORIGHT}
		local coords = {{28, 2}, {252, 2}, {28, 102}, {252, 102}}
		local n = c.pnum
		local front = n%2

		if not n return end	-- ????

		-- HP BAR:
		local basex = coords[n][1]
		local basey = coords[n][2]
		for i = 1, leveltime >= starttime and hlen or startlen
			if p.mo.zhudshake and hudshake.value
				rx = N_RandomRange(-4, 4)
				ry = N_RandomRange(-4, 4)	-- shaker for damage
			end
			v.drawFill(basex + (i-1)*2 +rx, basey+ry, 1, 8, ((shieldcount and 185) or (redhp and i > hashp) and (128) or (i > hashp) and (16) or (215))|flagsets[n])
		end
		if p.mo.zshieldinvinc
			v.drawString(basex + (front and -26 or 40), basey+ry, (p.mo.zshieldinvinc/35).."."..G_TicsToCentiseconds(p.mo.zshieldinvinc)/10, flagsets[n]|V_GREENMAP|V_HUDTRANS)
		else
			v.drawString(basex + (front and -26 or 40), basey+ry, tostring(p.zhp), flagsets[n]|(redhp and V_REDMAP or V_BLUEMAP)|V_HUDTRANS)
		end

		if btlzone
			hud_zdispzonegauge(v, p, c)
		end
		return
	end

	-- display HP, or shield duration if it's going on.
	if not p.mo.zshieldinvinc
		v.drawString(160+rx, 2+ry, tostring(p.zhp), V_SNAPTOTOP|(redhp and V_REDMAP or V_BLUEMAP)|V_HUDTRANS, "center")
	else
		v.drawString(160+rx, 2+ry, (p.mo.zshieldinvinc/35).."."..G_TicsToCentiseconds(p.mo.zshieldinvinc)/10, V_SNAPTOTOP|V_GREENMAP|V_HUDTRANS, "center")
	end


	for i = 1, leveltime >= starttime and hlen or startlen

		if p.mo.zhudshake and hudshake.value
			rx = N_RandomRange(-4, 4)
			ry = N_RandomRange(-4, 4)	-- shaker for damage
		end

		for j = 1, 2
			local startcoord = j == 1 and 130 or 182
			local getx = startcoord - xmod*(j == 1 and 1 or -1)

			v.drawFill(getx+rx, ry+hstart - (i-1)*2, 8, 1, ((shieldcount and 185) or (redhp and i > hashp) and (128) or (i > hashp) and (16) or (215))|V_SNAPTOTOP)

			if not xrev
				xmod = $+1
				if i >= hlen/2
					xrev = true
				end
			else
				xmod = $-1
			end
		end
	end

	if not btlzone
		-- display score for normal zone:
		V_drawString(v, 160+rx, 16+ry, tostring(p.zscore), "MKFNT", V_SNAPTOTOP|V_HUDTRANS, "center", true)
		V_drawString(v, 160+rx, 32+ry, "SCORE", "MKFNT", V_SNAPTOTOP|V_HUDTRANS, "center")
	else
		-- otherwise, display the zone and the zone gauge
		hud_zdispzonegauge(v, p, c)
		V_drawString(v, 160+rx, 16+ry, z_getdec(p.kzonespeed), "MKFNT", V_SNAPTOTOP|V_HUDTRANS, "center", true)
		V_drawString(v, 160+rx, 32+ry, "ZONE", "MKFNT", V_SNAPTOTOP|V_HUDTRANS, "center")
	end
end

-- display contextual messages
local function hud_zdispmsg(v, p, c)

	local by = btlzone and 160 or 180
	local i = #zonemsg
	while i

		local tflag = 0
		if zonemsg[i][2] < 20
			local t = (20 - zonemsg[i][2])/2
			tflag = V_10TRANS*t
		end

		v.drawString(160, by, zonemsg[i][1], V_SNAPTOBOTTOM|tflag, "center")

		zonemsg[i][2] = $-1
		if not zonemsg[i][2]
			table.remove(zonemsg, i)
			i = $-1
			continue
		end
		by = $-10
		i = $-1
	end
end

-- display results after the race is over

-- Start of Score Alterations - Latius
-- name of all the stuff we'll add together
local valuenames = {
	"SCORE",
	"ZONE",
	"PERFECT ZONES", -- Added
	"BEST ZONE COMBO", -- Added
	"TOP SPEED",
	"LAPS",
	"PERFECT LAPS",
}

-- multipliers for each of the above
local mults = {
	1,
	1000,
	0, -- Added
	1000, -- Added
	100,
	100,
	0, -- Added
}

local ktching = false	-- have we played the score tally sound?

-- this is very ghetto but it works so don't judge me you cunt
local function hud_zdispresults(v, p, c)

	if splitscreen return end
	if not p.mo return end

	local timer = p.mo.zresults
	if not timer return end	-- uh what

	if timer > TICRATE*2
		local fade = v.cachePatch("ZFADE")
		local t = max(V_50TRANS, V_90TRANS - V_10TRANS*((min(50, timer-TICRATE*2))/4))
		v.draw(0, 0, fade, V_SNAPTOTOP|V_SNAPTOLEFT|t)

		v.draw(0, 200, v.cachePatch("ZONELOGO"), t|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	end

	if timer > TICRATE*4

		--V_drawString(v, 160, 10, "RESULTS", "MKFNT", V_SNAPTOTOP, "center", true)
		v.draw(160, 10, v.cachePatch("ZONERES"))

		if timer > TICRATE*8
			local waitstr
			if z_countplayers() == 1 and not p.mo.z_exit	-- playing solo
				waitstr = "PRESS CUSTOM 1 TO EXIT"
			elseif not p.mo.z_exit
				waitstr = "PRESS CUSTOM 1 TO CONFIRM"
			elseif not p.exiting
				waitstr = "WAITING FOR PLAYERS... (PRESS F12 TO SPECTATE)"
			elseif p.exiting
				waitstr = "EXITING..."
			end
			v.drawString(160, 160, waitstr, V_SNAPTOBOTTOM|V_GREENMAP, "center")
		end

		-- check for modified game
		if p.zmodified	-- let's be lenient on HP, if the player chooses to start with less, it's their wish
			if timer > TICRATE*5
				v.drawString(160, 30, "NO SCORE ON MODIFIED GAMES", V_REDMAP, "center")
			end
			return
		end

		local vals = {
			p.zscore,
			p.kzonespeed/FRACUNIT,
			p.perfectzones, -- Added
			p.bestzonecombo, -- Added
			FixedDiv(FixedMul(p.maxspeed or 0, 142371), mapheaderinfo[gamemap].mobj_scale)/FRACUNIT,
			p.zlaps,
			p.perfectlaps, -- Added
		}

		if timer > TICRATE*4 + (TICRATE/3) else return end

		-- Added this to show the final Zone Speed
		v.drawString(60, 30, "ZONE SPEED:", V_YELLOWMAP)
		v.drawString(250, 30, z_getzonename(p.kzonespeed/FRACUNIT), V_BLUEMAP, "right")

		for i = 1, #vals do
			if timer > TICRATE*4 + (TICRATE/3*(i+1)) else break end

			v.drawString(60, 40+i*10, valuenames[i], V_YELLOWMAP) -- Altered to take into account the extra lines
			if mults[i] -- If it has an effect on the final score
				v.drawString(250, 40+i*10, vals[i], 0, "right")
				if mults[i] > 1
					v.drawString(180, 40+i*10, mults[i].."x", V_YELLOWMAP)
				end
			else
				v.drawString(250, 40+i*10, vals[i], V_GREENMAP, "right") -- If it doesn't, show it in green
			end
		end

		if timer > TICRATE*7 else return end -- Altered to work with the extra values

		v.drawFill(60, 125, 190, 1, 0) -- Altered to work with the extra values


		if timer > TICRATE*7 + (TICRATE/3) else return end -- Added
		local sum = 0
		for i = 1, #vals
			sum = $ + vals[i]*mults[i]
		end
		local timersum = (timer-(TICRATE*7+(TICRATE/3)))* max(1, sum/(TICRATE*4))

		v.drawString(60, 130, "TOTAL", V_YELLOWMAP) -- Altered to work with the extra values
		if timersum >= sum then
			-- local extics = (timersum - sum) / 50
			v.drawString(250, 130, sum, V_GREENMAP, "right")
			if not ktching
				S_StartSound(nil, sfx_chchng, p)
				ktching = true
			end
			-- May add in a Ranking sustem which may or may not be based on either a fixed value, or a value determined for the level, using extics as the timer for that
		else
			v.drawString(250, 130, timersum, 0, "right") -- Altered to work with the extra values
			S_StartSound(nil, sfx_menu1, p)
			ktching = false
		end
	end
end

local function z_huddrawbattleranking(v, p, c)

	if splitscreen return end
	if not p.mo return end

	-- random stuff for taking damage:
	local rx, ry = 0, 0
	if p.mo.zhudshake and hudshake.value
		rx = N_RandomRange(-4, 4)
		ry = N_RandomRange(-4, 4)	-- shaker for damage
	end

	-- draw target zone:
	if not (zonetargetflash and zonetargetflash%10 >= 5)
		V_drawString(v, 8+rx, 50+ry, "TARGET: \n"..dynzone..".0", "MKFNT", V_SNAPTOLEFT|V_HUDTRANS)
	end

	-- draw top 8
	local plyrs = {}
	for pl in players.iterate
		if not pl.mo continue end
		plyrs[#plyrs+1] = pl
	end
	-- sort by highest zone:
	table.sort(plyrs, function(a, b)
						if a.mo.kzonefinish and b.mo.kzonefinish
							return (a.mo.kzonefinish > b.mo.kzonefinish)
						elseif a.mo.kzonefinish and not b.mo.kzonefinish
							return true
						elseif not a.mo.kzonefinish and b.mo.kzonefinish
							return false
						else
							return (a.kzonespeed or 0) > (b.kzonespeed or 0)
						end
					end)

	-- display result:
	local y = 70
	for i = 1, 6
		if not plyrs[i]
			break
		end

		if p.mo.zhudshake and hudshake.value
			rx = N_RandomRange(-4, 4)
			ry = N_RandomRange(-4, 4)	-- shaker for damage
		end

		v.draw(8+rx, y+ry, v.cachePatch(skins[plyrs[i].mo.skin].facerank), V_HUDTRANS|V_SNAPTOLEFT, v.getColormap(plyrs[i].mo.skin, plyrs[i].mo.color))
		-- this is me?
		if plyrs[i] == p
			v.draw(8+rx, y+ry, v.cachePatch("K_CHILI"..((leveltime/4)%8)+1), V_HUDTRANS|V_SNAPTOLEFT)
		end
		-- pos
		v.draw(3, ry+y+10, v.cachePatch("OPPRNK"..string.format("%02d", tostring(i))), V_HUDTRANS|V_SNAPTOLEFT)
		-- zone
		local str = z_getdec(plyrs[i].kzonespeed or 0)
		if (plyrs[i].kzonespeed or 0)/FRACUNIT >= dynzone
			str = "FINISHED"
		end
		V_drawString(v, 24+rx, ry+y+8, str, "MKFNT", V_HUDTRANS|V_SNAPTOLEFT)
		y = $+17
	end

	-- draw time out
	if (zonetimeout and zonetimeout%10 < 5)
		V_drawString(v, 160+rx, 60+ry, "ZONE TIMEOUT", "MKFNT", V_HUDTRANS, "center", true)
	end

end

local hitfeedload

hud.add(function(v, p, c)
	displayplayer = p
	if not zoneon or G_BattleGametype() return end

	if leveltime < starttime/8 return end
	-- kill other hud elements:
	hud.disable("minirankings")
	hud.disable("gametypeinfo")
	hud.disable("position")
	hud.disable("item")
	hud.disable("time")
	hud.disable("speedometer")
	hud.disable("freeplay")

	if p.pflags & PF_TIMEOVER
	and not btlzone
		hud_zdispresults(v, p, c)
		hud.disable("minimap")
	else
		hud_zdispzone(v, p, c)
		hud_zdisphpandscore(v, p, c)
		hud_zdispmsg(v, p, c)
		if btlzone
			z_huddrawbattleranking(v, p, c)
		end
	end
end)

print("\x82".."Wipezones "..wver)
